@@ -1,14 +1,32 @@
 package Config::Ini::OnDrugs;
 
+# TODO: option to only allow include if owner is the same
+
 use 5.010;
 use strict;
 use warnings;
 use Log::Any '$log';
 
+use Cwd qw(abs_path);
 use Exporter::Lite;
+use File::Slurp;
+use File::chdir;
+
 our @EXPORT_OK = qw();
 
-our $VERSION = '0.01'; # VERSION
+our $VERSION = '0.02'; # VERSION
+
+use constant {
+    COL_RAW     => 0,
+    COL_INCLUDE => 1, # 1 if line is from included file
+    COL_TYPE    => 2, # [BCDPS]
+
+    COL_D_NAME  => 3,
+    COL_D_ARGS  => 4,
+    COL_P_NAME  => 3,
+    COL_P_VALUE => 4,
+    COL_S_NAMES => 3,
+};
 
 our %SPEC;
 
@@ -40,22 +58,30 @@ our $re_D_arg   = qr/$re_quotedl | [^"\s]+/x;
 our $re_D_args  = qr/(?:$re_D_arg (\s+ $re_D_arg)*)/x;
 our $re_D       = qr/^\s*
                      [;#]
-                     !(?<name>\w+) (?: \s+ $re_D_args )?
+                     !(?<name>\w+) (?<args> \s+ $re_D_args )?
                      \s*$
                     /x;
 
+sub _fmtmsg {
+    my ($self, $msg, $prefix) = @_;
+    join("",
+         $prefix // "",
+         (defined($self->{_curfile}) ?
+              " in file `$self->{_curfile}`" : ""),
+         (defined($self->{_curline}) ?
+              " at line #$self->{_curline}" : ""),
+        ": $msg"
+    );
+}
+
 sub _dieline {
     my ($self, $msg) = @_;
-    die "Parse error".(defined($self->{_curline}) ?
-                           " line #$self->{_curline}" : "").
-        ": $msg";
+    die $self->_fmtmsg($msg, "Parse error");
 }
 
 sub _warnline {
     my ($self, $msg) = @_;
-    $log->warn("Parse error".(defined($self->{_curline}) ?
-                                  " line #$self->{_curline}" : "").
-                                      ": $msg");
+    $log->warn($self->_fmtmsg($msg), "Parse warning");
 }
 
 sub _parse_quoted {
@@ -99,45 +125,64 @@ sub _split_args {
     @args;
 }
 
-sub dir_array {
-    my ($self, $args) = @_;
-    state $sub = sub {
-        my ($self, $args, $line) = @_;
-        if ($line->[2] =~ /\S/) {
-            $self->_warnline("Nulling non-empty value: $line->[2]");
-        }
-        $line->[2] = undef;
-    };
-    $self->{_next_param_hooks} = [$sub, 50];
-}
+sub _include {
+    my ($self, $filename) = @_;
+    $log->tracef("Including %s ...", $filename);
+    my $absfilename = abs_path($filename);
+    (-f $filename) && $absfilename or
+        $self->_dieline("Can't load file $filename: not found (cwd=".
+                    Cwd::cwd().")");
+    $self->{_include_stack} //= [];
+    if ($absfilename ~~ @{$self->{_include_stack}}) {
+        $self->_dieline("Recursive include: $absfilename");
+    }
+    my $dir;
+    if ($filename =~ m!/!) {
+        $dir = $filename;
+        $dir =~ s!(.+)/(.*)!$1!;
+    } else {
+        $dir = "";
+    }
 
-sub dir_null {
-    my ($self, $args) = @_;
-    state $sub = sub {
-        my ($self, $args, $line) = @_;
-        if ($line->[2] =~ /\S/) {
-            $self->_warnline("Nulling non-empty value: $line->[2]");
-        }
-        $line->[2] = undef;
-    };
-    $self->{_next_param_hooks} = [$sub, 50];
+    my $ct;
+    eval { $ct = read_file($filename) };
+    my $eval_err = $@;
+    if ($eval_err) {
+        $self->_dieline("Can't load file $filename: $eval_err");
+    }
+    push @CWD, $dir if length($dir);
+    local $self->{_curline};
+    local $self->{_curfile} = $absfilename;
+    push @{$self->{_include_stack}}, $absfilename;
+    $self->{_include_level}++;
+    $self->_parse_raw($ct);
+    $self->{_include_level}--;
+    pop @{$self->{_include_stack}};
+
+    pop @CWD if length($dir);
 }
 
+sub dirmeta_include { {phase=>1} }
 sub dir_include {
     my ($self, $args) = @_;
+    my $filename = $args->[0];
+    $self->_include($filename);
 }
 
+sub dirmeta_expr { {phase=>2} }
 sub dir_expr {
     my ($self, $args) = @_;
 }
 
+sub dirmeta_merge { {phase=>2} }
 sub dir_merge {
     my ($self, $args) = @_;
 }
 
 # parse raw ini untuk an array of lines. each line is an arrayref [RAW_LINE,
-# TYPE, (type-specific data ...)]. TYPE is either "B" (blank line), "C"
-# (comment), "D" (directive), "S" (section), or "P" (parameter).
+# TYPE(*)?, (type-specific data ...)]. TYPE is either "B" (blank line), "C"
+# (comment), "D" (directive), "S" (section), or "P" (parameter). * suffix
+# indicates that line comes from an included file.
 #
 sub _parse_raw {
     my ($self, $raw) = @_;
@@ -145,19 +190,24 @@ sub _parse_raw {
     if (!ref($raw) || ref($raw) ne 'ARRAY') {
         $raw = [split /^/, $raw];
     }
-    my @lines;
 
     $self->{_curline} = 0;
 
+    if ($self->{_include_level} == 0) {
+        $self->{_lines} = [];
+    }
+
     my @section_hooks;
     my @param_hooks;
     for my $line0 (@$raw) {
         $self->{_curline}++;
-        my @line = ($line0);
+        my @line;
+        $line[COL_RAW] = $line0;
+        $line[COL_INCLUDE] = $self->{_include_level} > 0;
 
         if ($line0 !~ /\S/) {
 
-            push @line, "B";
+            $line[COL_TYPE] = "B";
 
         } elsif ($line0 =~ /^\s*[;#](.*)/) {
 
@@ -165,23 +215,34 @@ sub _parse_raw {
             if ($arg =~ /^!\w+(?:\s+|$)/) {
                 if ($line0 =~ $re_D) {
                     my $name = $+{name};
+                    my $args0 = $+{args} // "";
                     my $meth = "dir_$name";
                     unless ($self->can($meth)) {
                         $self->_dieline("Unknown directive: $name");
                     }
-                    my $args = $+{args} // "";
-                    my @args = $self->_split_args($args);
-                    push @line, "D", $name, \@args;
+                    my @args = $self->_split_args($args0);
+                    $line[COL_TYPE]   = "D";
+                    $line[COL_D_NAME] = $name;
+                    $line[COL_D_ARGS] = \@args;
+                    my $methmt = "dirmeta_$name";
+                    unless ($self->can($methmt)) {
+                        $self->_dieline("Directive doesn't have meta: $name");
+                    }
+                    my $meta = $self->$methmt($name);
+                    if ($meta->{phase} == 1) {
+                        $self->$meth(\@args);
+                    }
                 } else {
                     $self->_warnline("Invalid directive syntax");
                 }
             } else {
-                push @line, "C";
+                $line[COL_TYPE] = "C";
             }
 
         } elsif ($line0 =~ /^\s*\[(.*)\]/) {
 
-            push @line, "S", $1; # XXX
+            $line[COL_TYPE]    = "S";
+            $line[COL_S_NAMES] = [$1]; # XXX
 
         } elsif ($line0 =~ /=/) {
 
@@ -190,7 +251,9 @@ sub _parse_raw {
                     $self->_parse_quoted($+{name_quoted}) : $+{name_bare};
                 my $value = defined($+{value_quoted}) ?
                     $self->_parse_quoted($+{value_quoted}) : $+{value_bare};
-                push @line, "P", $name, $value;
+                $line[COL_TYPE]    = "P";
+                $line[COL_P_NAME]  = $name;
+                $line[COL_P_VALUE] = $value;
             } else {
                 $self->_dieline("Invalid parameter assignment syntax");
             }
@@ -200,13 +263,13 @@ sub _parse_raw {
             $self->_dieline("Unknown line: $line0");
         }
 
-        push @lines, \@line;
+        push @{$self->{_lines}}, \@line;
     }
 
     # clean parsing work variables
-    undef $self->{_curline};
-
-    $self->{_lines} = \@lines;
+    if ($self->{_include_level} == 0) {
+        undef $self->{_curline};
+    }
 }
 
 # parse array of lines into tree of section/params.
@@ -219,16 +282,20 @@ sub _parse_lines {
     $self->{_tree} = {};
 
     for my $line (@{$self->{_lines}}) {
-        my $type = $line->[1];
+        my $type = $line->[COL_TYPE];
         if ($type eq 'S') {
-            my $name = $line->[2];
-            $self->{_cursection} = $line->[2];
+            my $name = $line->[COL_S_NAMES];
+            $self->{_cursection} = $name->[0]; # XXX
         } elsif ($type eq 'D') {
-            my ($name, $args) = ($line->[2], $line->[3]);
-            my $meth = "dir_$name";
-            $self->$meth($args);
+            my ($name, $args) = ($line->[COL_D_NAME], $line->[COL_D_ARGS]);
+            my $methmt = "dirmeta_$name";
+            my $meta = $self->$methmt($name);
+            if ($meta->{phase} == 2) {
+                my $meth = "dir_$name";
+                $self->$meth($args);
+            }
         } elsif ($type eq 'P') {
-            my ($name, $value) = ($line->[2], $line->[3]);
+            my ($name, $value) = ($line->[COL_P_NAME], $line->[COL_P_VALUE]);
             $self->{_tree}{ $self->{_cursection} }{$name} = $value;
         }
     }
@@ -242,20 +309,26 @@ sub _parse_lines {
 }
 
 sub new {
-    my ($class, $raw, $opts) = @_;
-    $opts //= {};
+    my ($class, %args) = @_;
+
     my $self = bless {}, $class;
 
-    for my $k (keys %$opts) {
-        die "Unknown attribute: $k" unless
-            $k =~ /\A(default_section)\z/;
-        $self->{$k} = $opts->{$k};
+    for my $k (keys %args) {
+        die "Unknown argument: $k" unless
+            $k =~ /\A(file|str|default_section)\z/;
+        $self->{$k} = $args{$k};
     }
-
     $self->{default_section} //= 'DEFAULT';
+    $self->{_include_stack} = [];
 
-    if (defined $raw) {
-        $self->_parse_raw($raw);
+    if (defined($self->{file}) || defined($self->{str})) {
+        if (defined $self->{file}) {
+            $self->{_include_level} = -1;
+            $self->_include($self->{file});
+        } elsif (defined $self->{str}) {
+            $self->{_include_level} = 0;
+            $self->_parse_raw($self->{str});
+        }
         $self->_parse_lines;
     }
     $self;
@@ -273,14 +346,19 @@ sub get_value {
     $s->{$param};
 }
 
+sub as_tree {
+    my ($self) = @_;
+    $self->{_tree};
+}
+
 # procedural functions
 
-$SPEC{a} = {
-    summary => '',
-    description => '',
-    args => {
-    },
-};
+#$SPEC{a} = {
+#    summary => '',
+#    description => '',
+#    args => {
+#    },
+#};
 
 1;
 #ABSTRACT: Yet another INI reader/writer (round trip, includes, variables, nest)
@@ -294,14 +372,13 @@ Config::Ini::OnDrugs - Yet another INI reader/writer (round trip, includes, vari
 
 =head1 VERSION
 
-version 0.01
+version 0.02
 
 =head1 SYNOPSIS
 
  # oo interface
- use File::Slurp;
  use Config::Ini::OnDrugs;
- my $ini = Config::Ini::OnDrugs->new(scalar read_file("file.ini"));
+ my $ini = Config::Ini::OnDrugs->new(file => "file.ini");
  my $section = $ini->get_section("Section"); # a hashref of param=>values
  my $val = $ini->get_value("Section", "Parameter");
 
@@ -317,8 +394,8 @@ version 0.01
  # dump back as string
  $ini->as_string;
 
- # procedural interface
- use Config::Ini::OnDrugs qw(
+ # procedural interface, Config::IOD is a shorter alias
+ use Config::IOD qw(
      ini_get
      ini_get_section
      ini get_value
@@ -476,7 +553,8 @@ Quoting is done with the double-quote (L<">) character. Known escapes are \',
 (backspace), \a (bell), \0, octal form ("\0377"), hex form ("\xff") and wide-hex
 form (\x{263a}).
 
-Quoting is allowed for section name in section line,
+Quoting is allowed for section name in section line and for parameter name and
+value in parameter line.
 
 =head2 Includes
 
@@ -563,15 +641,6 @@ Supported by Config::IniFiles. In Ini::OD, use multiple assignment:
 
 None are exported by default, but they are exportable.
 
-=head2 a(%args) -> [STATUS_CODE, ERR_MSG, RESULT]
-
-
-Returns a 3-element arrayref. STATUS_CODE is 200 on success, or an error code
-between 3xx-5xx (just like in HTTP). ERR_MSG is a string containing error
-message, RESULT is the actual result.
-
-No known arguments at this time.
-
 =head1 FAQ
 
 =head2 Why use INI format for configuration files?
